class KeyCloakResourceOwnerPassword:
    def __init__(self, ze_http_client, dd_url, token_url, username, password, idp_client_id, idp_client_secret,
                 verify_ssl=True, proxies=None) -> None:
        self.ze_http_client = ze_http_client
        self.verify_ssl = verify_ssl
        self.username = username
        self.password = password
        self.proxies = proxies
        self.token_url = token_url
        # http://localhost:8080/auth/realms/zema/protocol/openid-connect/token
        self.idp_client_id = idp_client_id

    def authenticate(self):
        headers = {
            'content-type': 'application/x-www-form-urlencoded'
        }

        payload = {'username': self.username,
                   'password': self.password,
                   'client_id': self.idp_client_id,
                   'grant_type': 'password'
                   }

        response = self.ze_http_client.post(self.token_url, data=payload, proxies=self.ze_http_client.proxies,
                                            headers=headers, verify=self.verify_ssl)

        tokens = response.json()
        return tokens["access_token"]



import json
from urllib.parse import parse_qs, urlparse

import requests


class OktaAuthorizationCode:
    def __init__(self, ze_http_client, dd_url, okta_issuer_url, username, password, verify_ssl=True,
                 proxies=None) -> None:
        self.dd_authorize_url = dd_url + "/oauth2/authorize"
        self.ze_http_client = ze_http_client
        self.verify_ssl = verify_ssl
        self.username = username
        self.password = password
        self.proxies = proxies

        # get the okta hostname?
        url = urlparse(okta_issuer_url)
        if url.scheme is not None and url.netloc is not None:
            self.okta_hostname_url = url.scheme + '://' + url.hostname
        else:
            raise ValueError('Invalid URL: ' + okta_issuer_url)

        self.okta_authentication_url = self.okta_hostname_url + '/api/v1/authn'

    def authenticate(self):

        payload = {
            'username': self.username,
            'password': self.password
        }

        headers = {
            'Accept': 'application/json',
            'Content-Type': 'application/json'
        }
        response = self.ze_http_client.post(self.okta_authentication_url, data=json.dumps(payload),
                                            proxies=self.ze_http_client.proxies,
                                            headers=headers)

        # get sessionToken
        session_token = response.json()['sessionToken']
        if session_token is None:
            raise ValueError('Invalid session token')

        # Get authorization code
        auth_init = self.ze_http_client.get(self.dd_authorize_url)
        auth_url = auth_init.url + f"&sessionToken={session_token}"
        code = self.ze_http_client.get(auth_url)



'''
Created on May 30, 2022

@author: jevan
'''
import requests, json, urllib.parse

class GenericResourceOwnerPasswordFlow:
    
    def __init__(self, username, password, verify_ssl=True, proxies=None, auth=None) -> None:
        self.token_url = auth['idp_token_url'] if 'idp_token_url' in auth else None
        self.client_id = auth['idp_client_id'] if 'idp_client_id' in auth else None
        self.client_secret = auth['idp_client_secret'] if 'idp_client_secret' in auth else None
        self.token_key = auth['idp_token_key'] if 'idp_token_key' in auth else 'id_token'
        self.scope = auth['idp_scope'] if 'idp_scope' in auth else 'openid'
        self.payload = {'grant_type': 'password', 'scope': self.scope, 'username': username, 'password': password}
        
        self.verify_ssl = verify_ssl
        self.proxies = proxies

    def get_bearer_token(self):
        token_response = requests.post(self.token_url, 
                               data=self.payload, 
                               verify=self.verify_ssl, 
                               allow_redirects=False, 
                               auth=(urllib.parse.quote(self.client_id), self.client_secret))

        tokens = json.loads(token_response.text)

        return tokens[self.token_key]



import base64
import json
from urllib.parse import parse_qs, urlparse

import requests
from bs4 import BeautifulSoup


class PingResourceOwnerPassword:
    def __init__(self, ze_http_client, dd_url, ping_url, username, password, idp_client_id, idp_client_secret,
                 verify_ssl=True, proxies=None) -> None:
        self.dd_authorize_url = dd_url + "/oauth2/authorize"
        self.ze_http_client = ze_http_client
        self.verify_ssl = verify_ssl
        self.username = username
        self.password = password
        self.proxies = proxies
        self.ping_token_url = ping_url + "/as/token.oauth2"
        self.idp_client_id = idp_client_id
        self.idp_client_secret = idp_client_secret

    def authenticate(self):
        code = f'{self.idp_client_id}:{self.idp_client_secret}'
        encoded = base64.b64encode(code.encode('ascii')).decode('ascii')

        headers = {
            'authorization': f'Basic {encoded}',
            'content-type': 'application/x-www-form-urlencoded'
        }

        payload = {'username': self.username,
                   'password': self.password,
                   'grant_type': 'password',
                   'scope': 'openid'}

        response = self.ze_http_client.post(self.ping_token_url, data=payload, proxies=self.ze_http_client.proxies,
                                            headers=headers, verify=self.verify_ssl)

        tokens = response.json()
        return tokens["access_token"]



"""
Created on May 8, 2019

@copyright: ZE
@precondition: ZEMA OData license
@version: 1.0.3
@author: ZE
"""

import requests
import json
import logging

class ZeHttpClient:
    """
    A Http client that interacts with ZE's REST and OData services
    """

    def __init__(self, url, proxies=None, headers=None, verify_ssl=True):
        self.url = url
        self.debug = False
        self.session = requests.Session()
        self.is_odata = True
        self.odata_version = -1
        if url.find('/api/v1') > 0:
            self.is_odata = False

        self.proxies = proxies
        self.headers = headers
        self.verify_ssl = verify_ssl
        self.logger = logging.getLogger(self.__module__)

    def enable_debug(self):
        self.debug = True

    def disable_debug(self):
        self.debug = False

    def login(self, username, password, client_id):
        if self.is_odata:
            self.session.auth = (username + '@' + client_id, password)
            response = self.session.get(self.url, proxies=self.proxies, verify=self.verify_ssl)
            if response.status_code != 200:
                raise Exception(
                    'Could not login. Http Code: {}, Message: {}'.format(response.status_code, response.text))
        else:
            response = self.session.post(self.url + '/login',
                                         data={'username': username, 'password': password, 'client': client_id},
                                         proxies=self.proxies, verify=self.verify_ssl)
            if response.status_code != 200:
                raise Exception(
                    'Could not login. Http Code: {}, Message: {}'.format(response.status_code, response.text))

            text = json.loads(response.text)
            if text['status'] != 'SUCCESS':
                raise Exception('Could not login. Message: {}'.format(text['message']))

    def get_resource(self, uri, payload=None):
        return self.get(self.url + uri, payload)

    def get(self, url, payload=None):
        response = self.session.get(url, params=payload, proxies=self.proxies, headers=self.headers,
                                    verify=self.verify_ssl)
        self.__check_response(response)
        if self.debug and self.logger.isEnabledFor(logging.DEBUG):
            self.__print_response(response)
        return response

    def post_resource(self, uri, payload):
        return self.post(self.url + uri, data=payload, proxies=self.proxies, headers=self.headers,
                         verify=self.verify_ssl)

    def post(self, url, data, proxies=None, headers=None, verify=True):
        response = self.session.post(url, data=data, proxies=proxies, headers=headers,
                                     verify=verify)
        self.__check_response(response)
        if self.debug and self.logger.isEnabledFor(logging.DEBUG):
            self.__print_response(response)
        return response

    def put_resource(self, uri, payload):
        return self.put(self.url + uri, data=payload, proxies=self.proxies, headers=self.headers,
                        verify=self.verify_ssl)

    def put(self, url, data, proxies=None, headers=None, verify=True):
        response = self.session.put(url, data=data, proxies=proxies, headers=headers,
                                    verify=verify)
        self.__check_response(response)
        if self.debug and self.logger.isEnabledFor(logging.DEBUG):
            self.__print_response(response)
        return response

    def patch_resource(self, uri, payload):
        return self.patch(self.url + uri, data=payload, proxies=self.proxies, headers=self.headers,
                          verify=self.verify_ssl)

    def patch(self, url, data, proxies=None, headers=None, verify=True):
        response = self.session.patch(url, data=data, proxies=proxies, headers=headers,
                                      verify=verify)
        self.__check_response(response)
        if self.debug and self.logger.isEnabledFor(logging.DEBUG):
            self.__print_response(response)
        return response

    def close(self):
        self.session.close()

    def __print_response(self, response):
        self.logger.debug('Request URL => {}'.format(response.url))
        self.logger.debug('Response Code => {}'.format(response.status_code))
        # may fail with encoding
        try:
            self.logger.debug('Response => {}'.format(response.text))
        except:
            self.logger.debug('Response => {...}')

    def __check_response(self, response):
        if response.status_code > 299:
            raise Exception(
                'Error encountered. Http Code: {}, Message: {}, Url: {}'.format(response.status_code, response.text,
                                                                                response.url))



import json
from urllib.parse import parse_qs, urlparse

import requests
from bs4 import BeautifulSoup


class PingAuthorizationCode:
    def __init__(self, ze_http_client, dd_url, ping_url, username, password, verify_ssl=True,
                 proxies=None) -> None:
        self.dd_authorize_url = dd_url + "/oauth2/authorize"
        self.ze_http_client = ze_http_client
        self.verify_ssl = verify_ssl
        self.username = username
        self.password = password
        self.proxies = proxies
        self.ping_hostname_url = ping_url

    def authenticate(self):
        # initialize the dd_authorize_url to get more settings of DD
        auth_init = self.ze_http_client.get(self.dd_authorize_url)
        # auth_url = auth_init.url

        doc = BeautifulSoup(auth_init.text, 'html.parser')
        form = doc.find('form')
        action = form.get('action')

        port_url = self.ping_hostname_url + action

        payload = {
            'pf.username': self.username,
            'pf.pass': self.password,
            # 'pf.adapterId': 'HTMLFormAdapter',
        }

        response = self.ze_http_client.post(port_url, data=payload,
                                            proxies=self.proxies, verify=self.verify_ssl)





import requests
from urllib.parse import parse_qs, urlparse


class AdfsBearerToken:

    def __init__(self, adfs_hostname_url, dd_url, username, password, verify_ssl=True, proxies=None) -> None:
        adfs_token_url = adfs_hostname_url + "/adfs/oauth2/token"
        adfs_authorize_url = adfs_hostname_url + "/adfs/oauth2/authorize"

        self.dd_authorize_url = dd_url + "/oauth2/authorize"

        # initialize the dd_authorize_url to get more settings of DD
        auth_init = requests.get(self.dd_authorize_url, verify=verify_ssl, proxies=proxies)
        auth_url = auth_init.url
        query = parse_qs(urlparse(auth_url).query)

        if len(query) == 0:
            raise ValueError(f'Could not init the connection to: {auth_url}')

        # Extract the params from the response redirect URL
        dd_client_id = query['client_id'][0]
        dd_redirect_uri = query['redirect_uri'][0]
        dd_resource = query['resource'][0]

        self.end_point = (f"{adfs_authorize_url}?" +
                          "response_type=code&" +
                          "scope=openid&" +
                          f"client_id={dd_client_id}&" +
                          f"redirect_uri={dd_redirect_uri}&" +
                          f"resource={dd_resource}")

        self.payload = (
                f"username={username}&" +
                f"password={password}&" +
                "AuthMethod=FormsAuthentication")
        self.dd_url = dd_url
        self.verify_ssl = verify_ssl
        self.proxies = proxies

    def get_bearer_token(self):
        response = requests.request("POST", self.end_point, data=self.payload, verify=self.verify_ssl,
                                    proxies=self.proxies)

        tokens = response.json()
        return tokens["data"]["token"]



from zeclient.http_client import ZeHttpClient
from zeclient.oauth2.generic_resource_owner_password_flow import GenericResourceOwnerPasswordFlow
from zeclient.oauth2.adfs_authorization_code import AdfsAuthorizationCode
from zeclient.oauth2.oauth2_flow import OAuth2Flow
from http.client import HTTPConnection
import logging

from zeclient.oauth2.okta_authorization_code import OktaAuthorizationCode
from zeclient.oauth2.ping_authorization_code import PingAuthorizationCode


class ZemaAuth:

    def __init__(self, url, username, password, client_id, enable_debug=False, proxies=None, auth=None) -> None:
        self.auth_url = url.replace('/odata4', '/odata')
        self.url = url
        self.username = username
        self.password = password
        self.client_id = client_id
        self.enable_debug = enable_debug
        self.proxies = proxies
        self.auth = auth
        self.verify_ssl = True
        self.logger = logging.getLogger(self.__module__)
        if self.auth is not None:
            self.verify_ssl = self.auth['verify_ssl'] if 'verify_ssl' in self.auth else True
            self.idp = self.auth['idp_type'] if 'idp_type' in self.auth else None
            self.idp_url = self.auth['idp_url'] if 'idp_url' in self.auth else None
            self.idp_token_url = self.auth['idp_token_url'] if 'idp_token_url' in self.auth else None
            self.idp_client_id = self.auth['idp_client_id'] if 'idp_client_id' in self.auth else None
            self.idp_client_secret = self.auth['idp_client_secret'] if 'idp_client_secret' in self.auth else None
            self.oauth2_flow = self.auth['oauth2_flow'] if 'oauth2_flow' in self.auth else None
            self.domain = self.auth['domain'] if 'domain' in self.auth else None
            if self.domain is not None:
                self.username = self.username + '@' + self.domain
        if self.enable_debug:
            HTTPConnection.debuglevel = 1

    def authenticate(self):

        self.enable_debug and self.logger.debug('Authenticating ' + self.username + " against " + self.auth_url)

        httpclient = None

        if self.auth is not None and self.oauth2_flow is not None:
            self.enable_debug and self.logger.debug(f'Authenticating using the provided oauth2_flow: {self.oauth2_flow}')
            if self.oauth2_flow is OAuth2Flow.ResourceOwnerPasswordCredentialsGrantType:
                resource_owner_flow = GenericResourceOwnerPasswordFlow(self.username,
                                                                       self.password,
                                                                       verify_ssl=self.verify_ssl,
                                                                       proxies=self.proxies,
                                                                       auth=self.auth)
                headers = {
                        'Authorization': f'Bearer {resource_owner_flow.get_bearer_token()}'
                }
                httpclient = ZeHttpClient(self.url, self.proxies, headers=headers, verify_ssl=self.verify_ssl)

            elif self.oauth2_flow is OAuth2Flow.AuthorizationCodeGrantType:                
                # ADFS
                if 'adfs' == self.idp:
                    httpclient = ZeHttpClient(self.url, self.proxies, verify_ssl=self.verify_ssl)
                    adfs_authorization_code = AdfsAuthorizationCode(httpclient,
                                                                    dd_url=self.auth_url,
                                                                    username=self.username,
                                                                    password=self.password,
                                                                    verify_ssl=self.verify_ssl)
                    adfs_authorization_code.authenticate()
                # OKTA
                elif 'okta' == self.idp:
                    httpclient = ZeHttpClient(self.url, self.proxies, verify_ssl=self.verify_ssl)
                    okta_authorization_code = OktaAuthorizationCode(httpclient,
                                                                    dd_url=self.auth_url,
                                                                    okta_issuer_url=self.idp_url,
                                                                    username=self.username,
                                                                    password=self.password,
                                                                    verify_ssl=self.verify_ssl)
                    okta_authorization_code.authenticate()
                # Ping Federate
                elif 'ping' == self.idp:
                    httpclient = ZeHttpClient(self.url, self.proxies, verify_ssl=self.verify_ssl)
                    ping_authorization_code = PingAuthorizationCode(httpclient,
                                                                    dd_url=self.auth_url,
                                                                    ping_url=self.idp_url,
                                                                    username=self.username,
                                                                    password=self.password,
                                                                    verify_ssl=self.verify_ssl,
                                                                    proxies=self.proxies)
                    ping_authorization_code.authenticate()
                # Key Cloak
                elif 'keycloak' == self.idp:
                    raise ValueError(f'The selected OAuth2 flow is not supported {self.oauth2_flow} for Keycloak')
                else:
                    raise ValueError(f'The IDP - {self.idp} is currently not supported')
            else:
                raise ValueError(f'The selected OAuth2 flow is not supported {self.oauth2_flow}')
            
        if httpclient is None:
            httpclient = ZeHttpClient(self.url, self.proxies, verify_ssl=self.verify_ssl)
            httpclient.login(self.username, self.password, self.client_id)

        if self.enable_debug:
            httpclient.enable_debug()

        return httpclient



'''
Utility functions for output

Created on Jan 03, 2019

@author: ZE
'''

from datetime import datetime
import pandas as pd

def print_profile_result(result):
    ending = ', '
    print('Date', end = ending)
    for i in range(0, len(result.header)):
        if (i == len(result.header) - 1):
            ending = '\n'
        
        if (result.header[i].caption != None):
            print(result.header[i].option, end = ending)
        elif (result.header[i].formula != None):
            print(result.header[i].formula, end = ending)
        elif (result.header[i].observation != None):
            print('%s' % (result.header[i].observation), end = ending)
        else:
            print(result.header[i].attribute[0].caption, end = ending)

    for row in range(0, len(result.data)):
        ending = ', '
        print(result.data[row].formattedDateString, end = ending)
        for col in range(0, len(result.data[row].result)):
            if (col == len(result.data[row].result) - 1):
                ending = '\n'
            print(result.data[row].result[col], end = ending)

def profile_result_to_df(result, raw=False):
    columns = []
    columns.append('Date')
    isHour = False
    isHourMinute = False
    if result.data[0].hour is not None:
        isHour = True
        columns.append('Hour')
    if result.data[0].minute is not None:
        isHourMinute = True
        columns.append('Minute')
        
    for i in range(0, len(result.header)):
        if result.header[i].caption is not None:
            columns.append(result.header[i].caption)
        elif raw and result.header[i].observation is not None:
            columns.append(result.header[i].observation)
        elif result.header[i].series is not None:
            columns.append(result.header[i].series)
        elif result.header[i].observation is not None:
            columns.append(result.header[i].observation)
        elif (result.header[i].formula is not None):
            columns.append(result.header[i].formula)
        else:
            columns.append(result.header[i].attribute[0].caption)

    df = {}
    for c in columns:
        df[c] = []
        
    for row in range(0, len(result.data)):
        i = 0
        df[columns[i]].append(datetime.strptime(result.data[row].date, "%m/%d/%Y").date())
        if isHour:
            i += 1
            if result.data[row].hour is not None:
                df[columns[i]].append(int(result.data[row].hour))
            else:
                df[columns[i]].append(None)
        if isHourMinute:
            i += 1
            if result.data[row].minute is not None:
                df[columns[i]].append(int(result.data[row].minute))
            else:
                df[columns[i]].append(None)
                
        for col in range(0, len(result.data[row].result)):
            i += 1
            df[columns[i]].append(result.data[row].result[col])
            
    return pd.DataFrame.from_dict(df)
    
def print_curve_bean(result):
    for c in result:
        print('ID: %d' % c.id)
        print('Name: %s' % c.name)
        print('Group name: %s' % c.groupName)
        print('Data type: %s' % c.dataType)
        print()
    
def curve_bean_to_df(result):
    df = {}
    columns = ['user', 'group', 'name', 'id', 'class', 'profile_name', 'config_name', 'curve_type', 'granularity', 'data_type']
    for c in columns:
        df[c] = []
    
    names = {}
    
    for b in result:                
        for p in b.profiles:
            if len(p.observations) == 0:
                df[columns[0]].append(b.owner)
                df[columns[1]].append(b.groupName)
                df[columns[2]].append(b.name)
                df[columns[3]].append(b.id)
                df[columns[4]].append(b.dataType)
                df[columns[5]].append(p.profileName)
                df[columns[6]].append(p.configName)
                df[columns[7]].append('')
                df[columns[8]].append(b.granularity)
                df[columns[9]].append('')
            else:
                for o in p.observations:
                    df[columns[0]].append(b.owner)
                    df[columns[1]].append(b.groupName)
                    df[columns[2]].append(b.name)
                    df[columns[3]].append(b.id)
                    df[columns[4]].append(b.dataType)
                    df[columns[5]].append(p.profileName)
                    df[columns[6]].append(p.configName)
                    df[columns[7]].append(o.curveType)
                    df[columns[8]].append(o.granularity)
                    df[columns[9]].append(o.dataType)
        
        if len(b.properties) > 0:
            for pp in b.properties:
                if pp.label in names:
                    names[pp.label] += 1
                else:
                    names[pp.label] = 1
    
    result_len = len(result)
    valid_names = []
    for n in names:
        if names[n] == result_len:
            valid_names.append(n)
            df[n] = []
    
    # only include properties assigned to all available curves
    if len(valid_names) > 0:
        for b in result:
            for pp in b.properties:
                if pp.label in valid_names:
                    for p in b.profiles:
                        if len(p.observations) == 0:
                            df[pp.label].append(pp.value)
                        else:
                            for o in p.observations:
                                df[pp.label].append(pp.value)
        
    return pd.DataFrame.from_dict(df)
    
def print_curve_data_bean(result):
    for r in result:
        print('Curve name: %s' % r.curveName)
        print('Opr date: %s' % r.oprDate)
        print('Curve type: %s' % r.curveType)
        print('Contract year: %d' % r.contractYear)
        print('Contract month: %d' % r.contractMonth)
        print('Value: %s' % r.value)
        print()

def print_forward_curve_data(result):
    for gd in result.groupedData:
        print('Key: %s' % gd.key)
        for fd in gd.data:
            print('\tCurve name: %s' % fd.curveName)
            print('\tOpr date: %s' % fd.oprDate)
            print('\tCurve type: %s' % fd.curveType)
            print('\tContract year: %d' % fd.contractYear)
            print('\tContract month: %d' % fd.contractMonth)
            print('\tValue: %s' % fd.value)
            print()

def forward_curve_to_df(result):
    df = {}
    columns = ['name', 'opr_date', 'contract_start', 'contract_end', 'contract_code', 'contract_year', 'type', 'value', 'date_modified']
    for c in columns:
        df[c] = []

    for gd in result.groupedData:
        for e in gd.data:
            df[columns[0]].append(e.curveName)
            df[columns[1]].append(e.oprDate.date())
            df[columns[2]].append(e.contractStart.date())
            df[columns[3]].append(e.contractEnd.date())
            df[columns[4]].append(e.contractMonth)
            df[columns[5]].append(e.contractYear)
            df[columns[6]].append(e.curveType)
            df[columns[7]].append(e.value)
            df[columns[8]].append(e.lastUpdated)
            
    return pd.DataFrame.from_dict(df)

def time_series_curve_to_df(result):
    df = {}
    columns = ['name', 'opr_date', 'opr_hour', 'opr_minute', 'type', 'value', 'date_modified']
    for c in columns:
        df[c] = []

    for gd in result.groupedData:
        for e in gd.data:
            df[columns[0]].append(e.curveName)
            df[columns[1]].append(e.oprDate.date())
            df[columns[2]].append(e.oprHour)
            df[columns[3]].append(e.oprMinute)
            df[columns[4]].append(e.curveType)
            df[columns[5]].append(e.value)
            df[columns[6]].append(e.lastUpdated)
            
    return pd.DataFrame.from_dict(df)

def options_curve_to_df(result):
    df = {}
    columns = ['name', 'opr_date', 'contract_start', 'contract_end', 'contract_code', 'contract_year', 'type',
               'put_call', 'level_type', 'level_value', 'strip_unit', 'spread_length', 'contract_start_2',
               'contract_end_2', 'contract_month_2', 'contract_year_2', 'value', 'date_modified']
    for c in columns:
        df[c] = []

    for gd in result.groupedData:
        for e in gd.curveData:
            df[columns[0]].append(e.curveName)
            df[columns[1]].append(e.oprDate.date())
            df[columns[2]].append(e.contractStart.date())
            df[columns[3]].append(e.contractEnd.date())
            df[columns[4]].append(e.contractMonth)
            df[columns[5]].append(e.contractYear)
            df[columns[6]].append(e.curveTypeName)
            df[columns[7]].append(e.putCall)
            df[columns[8]].append(e.levelType)
            df[columns[9]].append(e.levelValue)
            df[columns[10]].append(e.stripUnit)
            df[columns[11]].append(e.spreadLength)
            df[columns[12]].append(e.contractStart2.date())
            df[columns[13]].append(e.contractEnd2.date())
            df[columns[14]].append(e.contractMonth2)
            df[columns[15]].append(e.contractYear2)
            df[columns[16]].append(e.value)
            df[columns[17]].append(e.lastUpdated)
            
    return pd.DataFrame.from_dict(df)

def insert_update_curve_data_result_to_df(result):
    df = {}
    columns = ['curve_name', 'curve_id', 'opr_date', 'records_inserted', 'records_updated', 'records_deleted', 'valid']
    for c in columns:
        df[c] = [] 
        
    for e in result:
        df[columns[0]].append(e.curveName)
        df[columns[1]].append(e.curveId)
        df[columns[2]].append(e.oprDate)
        df[columns[3]].append(e.numRecordsInserted)
        df[columns[4]].append(e.numRecordsUpdated)
        df[columns[5]].append(e.numRecordsDeleted)
        df[columns[6]].append(e.valid)
    
    return pd.DataFrame.from_dict(df)
        
def print_time_series_curve_data(result):
    for gd in result.groupedData:
        print('Key: %s' % gd.key)
        for fd in gd.data:
            print('\tCurve name: %s' % fd.curveName)
            print('\tOpr date: %s' % fd.oprDate)
            if (fd.oprHour != None):
                print('\tOpr hour: %s' % fd.oprHour)
            print('\tCurve type: %s' % fd.curveType)
            print('\tValue: %s' % fd.value)
            print()

def print_options_curve_data(result):
    for gd in result.groupedData:
        print('Key: %s' % gd.key)
        for fd in gd.curveData:
            print('\tCurve name: %s' % fd.curveName)
            print('\tOpr date: %s' % fd.oprDate)
            print('\tCurve type: %s' % fd.curveTypeName)
            print('\tContract year: %d' % fd.contractYear)
            print('\tContract month: %d' % fd.contractMonth)
            print('\tPut / Call: %s' % fd.putCall)
            print('\tLevel type: %s' % fd.levelType)
            print('\tLevel value: %s' % fd.levelValue)
            print('\tStrip unit: %s' % fd.stripUnit)
            print('\tSpread length: %d' % fd.spreadLength)
            print('\tContract year 2: %d' % fd.contractYear2)
            print('\tContract month 2: %d' % fd.contractMonth2)
            print('\tValue: %s' % fd.value)
            print()



'''
ZE SOAP Web Services Clinet

Created on May 21, 2019

@copyright: ZE
@precondition: ZEMA Data Direct license
@version: 1.0.3
@author: ZE
'''

import datetime
from zeclient.soap_proxy import SoapProxy
import zeclient.soap_util as util

class SoapClient:
    '''
    A client for ZE SOAP services
    '''
    def __init__(self,url,username,password,client_id,enable_debug=False,proxies=None):
        if url.endswith('/'):
            url = url[:-1]

        if url.find('/services/dd') < 0:
            url = url + '/services/dd'
                        
        if url.find('?wsdl') < 0:
            url = url + '?wsdl'

        self.wsdl = url        
        self.proxy = SoapProxy(url, username, password, client_id,proxies)
        self.debug = enable_debug
        
        if enable_debug:
            print('SOAP authenticating ' + username + " against " + url)

    def get_profile(self, username, group, name, config=None, effective_date=None):
        param = {}
        param['profileOwner'] = username
        param['profileGroup'] = group
        param['profileName'] = name
        if config is not None:
            param['configName'] = config
            
        if effective_date is not None:
            param['effectiveDate'] = effective_date.isoformat()
        
        result = self.proxy.execute_profile(param)
        return util.profile_result_to_df(result)
    
    def get_report(self, datasource, report, start_date, end_date, observations=None, filters=None):
        if not observations: # get all observations if it's empty
            lvs = self.proxy.retrieve_observations(datasource, report)
            observations = []
            for e in lvs:
                observations.append(e.value)
            
        dr = {}
        dr['options'] = {'dataOptions': {'precision': '5'},
                         'timelineOptions': {'startDate': self.__date_to_string(start_date),
                                             'endDate': self.__date_to_string(end_date)}
                        }
        dataseries = {'dataSource': datasource,
                      'dataReport': report,
                      'observation': observations.pop()
                     }
            
        attr = []
        if filters is not None:
            for key in filters:
                afilter = {'columnName': key}
                values = []
                t = type(filters[key])
                if t == datetime.date or t == datetime.datetime:
                    values.append(str(filters[key].isoformat()))
                elif t == bool:
                    values.append(str(filters[key]).lower())
                elif t == list:
                    values = filters[key]
                else:
                    values.append(filters[key])
                afilter['values'] = values
                attr.append(afilter)
        
        if attr:
            dataseries['attributes'] = attr
        
        dr['dataSeries'] = dataseries
        
        if observations:
            olist = []
            for o in observations:
                olist.append({'observation': o})
            dr['observations'] = olist
        
        if self.debug:
            print(dr)
            
        result = self.proxy.execute_report_query('Data Sheet', {'useDisplayName': 'false',
                                                                'dataRequest': dr})

        return util.profile_result_to_df(result, True)

    def find_curves(self, filters):
        g_filters =  filters['groups'] if 'groups' in filters else {}
        n_filters = filters['names'] if 'names' in filters else {}
        p_filters = filters['properties'] if 'properties' in filters else {}

        param = {}

        if g_filters:
            for op in g_filters:
                param['groupsStringMatchType'] = self.__get_string_match(op)
                t = type(g_filters[op])
                if t == list:
                    param['groups'] = g_filters[op]
                else:
                    param['groups'] = [g_filters[op]]
        
        if n_filters:
            for op in n_filters:
                param['namesStringMatchType'] = self.__get_string_match(op)
                t = type(n_filters[op])
                if t == list:
                    param['names'] = n_filters[op]
                else:
                    param['names'] = [n_filters[op]]
        
        if p_filters:
            properties = []
            for n in p_filters:
                properties.append({'label': n, 'value': p_filters[n]})
            param['properties'] = properties
        
        param['includeProperties'] = filters['includeProperties'] if 'includeProperties' in filters else False
        param['includeObservations'] = filters['includeObservations'] if 'includeObservations' in filters else False
        
        result = self.proxy.find_curves(param)

        return util.curve_bean_to_df(result)
            
    def get_forward_curve(self, start_date, end_date, name = None, property_filters = None):
        param = self.__get_curve_data_request(start_date, end_date, name, property_filters)
        
        result = self.proxy.get_forward_curve_data(param)
        
        return util.forward_curve_to_df(result)
    
    def get_timeseries_curve(self, start_date, end_date, name = None, property_filters = None):
        param = self.__get_curve_data_request(start_date, end_date, name, property_filters)
        
        result = self.proxy.get_time_series_curve_data(param)
        
        return util.time_series_curve_to_df(result)
    
    def get_options_curve(self, start_date, end_date, name = None, property_filters = None):
        param = self.__get_curve_data_request(start_date, end_date, name, property_filters)
        param['pivoted'] = False
        
        result = self.proxy.get_options_curve_data(param)
        
        return util.options_curve_to_df(result)
    
    def insert_update_curve_data(self, data):
        result = self.proxy.insert_update_curve_data(data)
        return util.insert_update_curve_data_result_to_df(result)
    
    def __date_to_string(self, d):
        return d.strftime("%m/%d/%Y")
    
    def __get_string_match(self, op):
        m = ''
        if op == 'startswith':
            m = 'STARTS_WITH'
        elif op == 'endswith':
            m = 'ENDS_WITH'
        elif op == 'equals':
            m = 'EQUALS'
        elif op == 'contains':
            m = 'CONTAIINS'
        else:
            raise Exception('Unknow match type - {}'.format(op))
        return m

    def __get_curve_data_request(self, start_date, end_date, name = None, property_filters = None):
        param = {}
        
        if name is not None:
            if type(name) == list:
                param['curveNames'] = name
            else:
                param['curveNames'] = [name]
            
        param['startDate'] = start_date.isoformat()
        param['endDate'] = end_date.isoformat()
        
        if property_filters is not None:
            properties = []
            for p in property_filters:
                vs = property_filters(p)
                t = type(vs)
                if t == list:
                    for v in vs:
                        properties.append({'label': p, 'value': v})
                else:
                    properties.append({'label': p, 'value': vs})
        
        param['includePropertiesInResponse'] = False
        param['updatedOnly'] = False
        param['pivoted'] = False

        return param
    



"""
ZE ZEMA Clinet

Created on May 28, 2019

@copyright: ZE
@precondition: ZEMA Data Direct license
@version: 1.0.3
@author: ZE
"""

from zeclient.odata_client import ODataClient
from zeclient.rest_client import RestClient

class ZemaClient:

    def __init__(self, datadirect_url, username, password, client_id, enable_debug=False, proxies=None, auth=None):
        self.odata_client = ODataClient(datadirect_url, username, password, client_id, enable_debug, proxies=proxies,
                                        auth=auth)
        self.rest_client = RestClient(datadirect_url, username, password, client_id, enable_debug, proxies=proxies,
                                      auth=auth)

    def get_profile(self, username, group, name, config=None, select=None, filters=None, top=None, order_by=None,
                    skip=None):
        return self.odata_client.get_profile(username, group, name, config, select, filters, top, order_by, skip)

    def get_report(self, datasource, report, select=None, filters=None, top=None, order_by=None, skip=None):
        return self.odata_client.get_report(datasource, report, select, filters, top, order_by, skip)

    def get_holidays(self, select=None, filters=None, top=None, order_by=None, skip=None):
        return self.odata_client.get_holidays(select, filters, top, order_by, skip)

    def find_profile_users(self, select=None, filters=None, top=None):
        return self.odata_client.find_profile_users(select, filters, top)

    def find_profile_groups(self, select=None, filters=None, top=None):
        return self.odata_client.find_profile_groups(select, filters, top)

    def find_profiles(self, select=None, filters=None, top=None):
        return self.odata_client.find_profiles(select, filters, top)
    
    def find_holiday_groups(self, select=None, filters=None, top=None):
        return self.odata_client.find_holiday_groups(select, filters, top)

    def find_linked_profiles(self, select=None, filters=None, top=None):
        return self.odata_client.find_linked_profiles(select, filters, top)

    def find_data_sources(self, select=None, filters=None, top=None):
        return self.odata_client.find_data_sources(select, filters, top)

    def find_reports(self, select=None, filters=None, top=None):
        return self.odata_client.find_reports(select, filters, top)

    def find_curve_groups(self, select=None, filters=None, top=None):
        return self.odata_client.find_curve_groups(select, filters, top)

    def find_curves(self, select=None, filters=None, top=None):
        return self.odata_client.find_curves(select, filters, top)

    def get_forward_curve(self, name=None, select=None, filters=None, top=None, order_by=None, skip=None):
        return self.odata_client.get_forward_curve(name, select, filters, top, order_by, skip)

    def get_timeseries_curve(self, name=None, select=None, filters=None, top=None, order_by=None, skip=None):
        return self.odata_client.get_timeseries_curve(name, select, filters, top, order_by, skip)

    def get_options_curve(self, name=None, select=None, filters=None, top=None, order_by=None, skip=None):
        return self.odata_client.get_options_curve(name, select, filters, top, order_by, skip)

    def get_forward_data_and_property(self, start_date, end_date, curve_filters, include_properties=None, top=None):
        column_map = {"name": "name",
                      "oprDate": "opr_date",
                      "contractMonth": "contract_month",
                      "contractYear": "contract_year",
                      "contractStart": "contract_start",
                      "contractEnd": "contract_end",
                      "type": "type",
                      "value": "value",
                      "dateModified": "date_modified",
                      "group": "group",
                      "contractGranularity": "contract_granularity"}
        pd = self.rest_client.get_forward_curve(start_date, end_date, curve_filters, include_properties)
        pd = pd.rename(columns=column_map)
        if top is not None:
            return pd.head(top)
        else:
            return pd

    def get_timeseries_data_and_property(self, start_date, end_date, curve_filters, include_properties=None, top=None):
        column_map = {"name": "name",
                      "oprDate": "opr_date",
                      "oprHour": "opr_hour",
                      "oprMinute": "opr_minute",
                      "type": "type",
                      "value": "value",
                      "group": "group",
                      "dateModified": "date_modified"}
        pd = self.rest_client.get_timeseries_curve(start_date, end_date, curve_filters, include_properties)
        pd = pd.rename(columns=column_map)
        if top is not None:
            return pd.head(top)
        else:
            return pd

    def get_options_data_and_property(self, start_date, end_date, curve_filters, include_properties=None, top=None):
        column_map = {'name': 'name',
                      'group': 'group',
                      'oprDate': 'opr_date',
                      'contractMonth': 'contract_code',
                      'contractYear': 'contract_year',
                      'contractStart': 'contract_start',
                      'contractEnd': 'contract_end',
                      'levelType': 'level_type',
                      'levelValue': 'level_value',
                      'putCall': 'call_put',
                      'spreadLength': 'spread_length',
                      'stripUnit': 'strip_unit',
                      'type': 'type',
                      'contractMonth2': 'contract_month_2',
                      'contractYear2': 'contract_year_2',
                      'contractStart2': 'contract_start_2',
                      'contractEnd2': 'contract_end_2',
                      'contractGranularity': 'contract_granularity',
                      'value': 'value',
                      'dateModified': 'date_modified'}
        pd = self.rest_client.get_options_curve(start_date, end_date, curve_filters, include_properties)
        pd = pd.rename(columns=column_map)
        if top is not None:
            return pd.head(top)
        else:
            return pd

    def upload_curve_data(self, group_name, curve_name, payload, partial_update=None):
        return self.rest_client.upload_curve_data(group_name, curve_name, payload, partial_update)

    def enable_debug(self):
        self.odata_client.enable_debug()
        self.rest_client.enable_debug()

    def disable_debug(self):
        self.odata_client.disable_debug()
        self.rest_client.disable_debug()

    def close(self):
        self.odata_client.close()
        self.rest_client.close()



"""
Created on May 3, 2019

@copyright: ZE
@precondition: ZEMA OData license
@version: 1.0.3
@author: ZE
"""

import json
import pandas as pd
import urllib
from datetime import date
from datetime import datetime

from zeclient.zema_auth import ZemaAuth


class ODataClient:
    """
    A client for ZE OData services
    """

    def __init__(self, url, username, password, client_id, enable_debug=False, proxies=None, auth=None):
        self.odata_version = 4
        if url.endswith('/'):
            url = url[:-1]

        if url.find('/odata') < 0:
            url = url + '/odata4'

        if url.find('/odata4') < 0:
            raise Exception('ZemaClient only supports OData version 4')

        if url.find('/v1') < 0:
            url = url + '/v1'

        self.__httpclient = None
        self.__zema_auth = ZemaAuth(url, username, password, client_id, enable_debug, proxies, auth)

    def get_profile(self, username, group, name, config=None, select=None, filters=None, top=None, order_by=None,
                    skip=None):
        profile_id = self.get_profile_entity_id(username, group, name, config)

        payload = self.__payload(select=select, filters=filters, top=top, order_by=order_by, skip=skip)

        response = self.__get_httpclient().get_resource('/' + profile_id, payload)

        return pd.DataFrame.from_dict(self.__values(response))

    def get_report(self, datasource, report, select=None, filters=None, top=None, order_by=None, skip=None):
        report_id = self.get_report_entity_id(datasource, report)

        payload = self.__payload(select=select, filters=filters, top=top, order_by=order_by, skip=skip)

        response = self.__get_httpclient().get_resource('/' + report_id, payload)

        return pd.DataFrame.from_dict(self.__values(response))

    def find_profile_users(self, select=None, filters=None, top=None):
        return self.__find('/profile_users', select=select, filters=filters, top=top)

    def find_profile_groups(self, select=None, filters=None, top=None):
        return self.__find('/profile_groups', select=select, filters=filters, top=top)

    def find_profiles(self, select=None, filters=None, top=None):
        return self.__find('/profiles', select=select, filters=filters, top=top)

    def find_linked_profiles(self, select=None, filters=None, top=None):
        return self.__find('/linked_profiles', select=select, filters=filters, top=top)

    def find_data_sources(self, select=None, filters=None, top=None):
        return self.__find('/data_sources', select=select, filters=filters, top=top)

    def find_reports(self, select=None, filters=None, top=None):
        return self.__find('/reports', select=select, filters=filters, top=top)

    def find_holiday_groups(self, select=None, filters=None, top=None):
        return self.__find('/holiday_groups', select=select, filters=filters, top=top)

    def get_holidays(self, select=None, filters=None, top=None, order_by=None, skip=None):
        payload = self.__payload(select=select, filters=filters, top=top, order_by=order_by, skip=skip)
        response = self.__get_httpclient().get_resource('/holidays', payload)
        return pd.DataFrame.from_dict(self.__values(response))

    def find_curve_groups(self, select=None, filters=None, top=None):
        return self.__find('/curve_groups', select=select, filters=filters, top=top)

    def find_curves(self, select=None, filters=None, top=None):
        return self.__find('/curves', select=select, filters=filters, top=top)

    def get_forward_curve(self, name=None, select=None, filters=None, top=None, order_by=None, skip=None):
        return self.__curve_data('/futures_curve_data', name, select, filters, top, order_by, skip)

    def get_timeseries_curve(self, name=None, select=None, filters=None, top=None, order_by=None, skip=None):
        return self.__curve_data('/time_series_curve_data', name, select, filters, top, order_by, skip)

    def get_options_curve(self, name=None, select=None, filters=None, top=None, order_by=None, skip=None):
        return self.__curve_data('/options_curve_data', name, select, filters, top, order_by, skip)

    def get_profile_entity_id(self, username, group, name, config=None):
        uri = None
        if group is None or group == '':
            group = 'NOGROUP'

        if config is None or config == '':
            uri = "/profiles(user='{}',group='{}',name='{}')".format(username, group, name)
        else:
            uri = "/linked_profiles(user='{}',template_group='{}',template='{}',name='{}')".format(username, group,
                                                                                                   name, config)
        
        response = self.__get_httpclient().get_resource(uri, {'$format': 'json'})
        json_object = json.loads(response.text)
        profile_id = None

        if 'd' not in json_object:  # odata 4
            profile_id = json_object['data_entity_id']
        else:
            profile_id = json_object['d']['data_entity_id']

        return profile_id

    def get_report_entity_id(self, datasource, report):
        uri = "/reports(source='{}',name='{}')".format(datasource, report)
        response = self.__get_httpclient().get_resource(uri, {'$format': 'json'})
        json_object = json.loads(response.text)
        report_id = None

        if 'd' not in json_object:  # odata 4
            report_id = json_object['data_entity_id']
        else:
            report_id = json_object['d']['data_entity_id']

        return report_id

    def enable_debug(self):
        if self.__httpclient is not None:
            self.__httpclient.enable_debug()

    def disable_debug(self):
        if self.__httpclient is not None:
            self.__httpclient.disable_debug()

    def close(self):
        if self.__httpclient is not None:
            self.__httpclient.close()
        
    def __get_httpclient(self):
        if self.__httpclient is None:
            self.__httpclient = self.__zema_auth.authenticate()
            
        return self.__httpclient

    def __find(self, entity_url, select=None, filters=None, top=None):
        payload = self.__payload(select=select, filters=filters, top=top)
        response = self.__get_httpclient().get_resource(entity_url, payload)
        return pd.DataFrame.from_dict(self.__values(response))

    def __curve_data(self, uri, name=None, select=None, filters=None, top=None, order_by=None, skip=None):
        if name is not None:
            if filters is None:
                filters = {}
            filters['name'] = {'op': 'eq', 'value': name}

        payload = self.__payload(select=select, filters=filters, top=top, order_by=order_by, skip=skip)
        response = self.__get_httpclient().get_resource(uri, payload)

        return pd.DataFrame.from_dict(self.__values(response))

    def __values(self, response):
        values = None
        json_object = json.loads(response.text)
        if self.odata_version == 4:
            values = json_object['value']
        else:
            values = json_object['d']['results']
            for e in values:
                del e['__metadata']

        return values

    def __payload(self, select=None, filters=None, top=None, order_by=None, skip=None):
        payload = {'$format': 'json'}
        if select is not None:
            payload["$select"] = ",".join(select) if type(select) == list else select

        filter_clause = None

        if filters is not None:
            all_columns = []
            for key in filters:
                afilter = filters[key]
                if type(afilter) == list:
                    for e in afilter:
                        self.__filter(all_columns, key, e)
                else:
                    self.__filter(all_columns, key, afilter)

            filter_clause = ('' if filter_clause is None else (filter_clause + ' and ')) + " and ".join(all_columns)

        if filter_clause is not None:
            payload["$filter"] = filter_clause.encode(encoding='utf-8')

        if top is not None:
            payload["$top"] = top

        if order_by is not None:
            payload["$orderby"] = order_by

        if skip is not None:
            payload["$skip"] = skip

        # by default, requests encode space to plus and this is incorrect
        payload = urllib.parse.urlencode(payload, quote_via=urllib.parse.quote)

        return payload

    def __filter(self, all_expressions, name, operation):
        if type(operation['value']) == list:
            expressions = []
            for e in operation['value']:
                expressions.append(self.__single_filter(name, {'op': operation['op'], 'value': e}))

            all_expressions.append('(' + ' or '.join(expressions) + ')')
        else:
            all_expressions.append(self.__single_filter(name, operation))

        return all_expressions

    def __single_filter(self, name, operation):
        e = operation['value']

        if len(operation['op']) > 2:
            if operation['op'] == 'startswith':
                return "startswith({},'{}') eq true".format(name, e)
            elif operation['op'] == 'endswith':
                return "endswith({},'{}') eq true".format(name, e)
            elif operation['op'] == 'contains':
                return "indexof({},'{}') ge 0".format(name, e)
            else:
                raise Exception('Unknown operator - ' + operation['op'])
        else:
            if type(e) == str:
                return name + ' ' + operation['op'] + ' ' + "'" + e + "'"
            elif type(e) == date:
                return name + ' ' + operation['op'] + ' ' + str(e.isoformat())
            elif type(e) == datetime:
                return name + ' ' + operation['op'] + ' ' + str(e.isoformat()) + 'Z' if e.tzinfo is None else ''
            else:
                return name + ' ' + operation['op'] + ' ' + str(e)



"""
Created on May 3, 2019

@copyright: ZE
@precondition: ZEMA license
@version: 1.0.3
@author: ZE
"""

import json
import pandas as pd
import datetime
import calendar

from zeclient.zema_auth import ZemaAuth


class RestClient:
    """
    A client for ZE REST services
    """

    def __init__(self, url, username, password, client_id, enable_debug=False, proxies=None, auth=None):
        if url.endswith('/'):
            url = url[:-1]

        if url.find('/api') < 0:
            url = url + '/api'

        if url.find('/v1') < 0:
            url = url + '/v1'

        self.__zema_auth = ZemaAuth(url, username, password, client_id, enable_debug, proxies, auth)
        self.__httpclient = None

    def get_report(self, data_source, report, start_date, end_date, select=None, filters=None):
        uri = '/data/sources/{}/reports/{}'.format(data_source, report)

        response = self.__get_httpclient().get_resource(uri, self.__payload(start_date, end_date, select, filters))
        result = self.__data(response)

        d = []
        for e in result:
            for columns in e['columns']:
                o = {}
                o['date'] = e['date']
                for c in columns:
                    o[c['name']] = c['value']
                d.append(o)

        return pd.DataFrame.from_dict(d)

    def get_result_set(self, data_source, report, filters=None):
        uri = "/resultset/sources/{}/reports/{}".format(data_source, report)        
        response = self.__get_httpclient().get_resource(uri, self.__payload(filters=filters))
        json_object = json.loads(response.text)
        series = json_object['data']['series']
        d = []
        for e in series:
            o = {}
            for c in e['column']:
                o[c['name']] = c['value']

            o['minDate'] = e['minDate']
            o['maxDate'] = e['maxDate']
            d.append(o)

        return pd.DataFrame.from_dict(d)

    def find_data_sources(self):
        uri = "/metadata/sources"
        return self.__find(uri)

    def find_reports(self, data_source):
        uri = "/metadata/sources/{}/reports".format(data_source)
        return self.__find(uri)

    def find_report_observations(self, data_source, report):
        uri = "/metadata/sources/{}/reports/{}/observations".format(data_source, report)
        return self.__find(uri)

    def find_report_attributes(self, data_source, report):
        uri = "/metadata/sources/{}/reports/{}/attributes".format(data_source, report)
        return self.__find(uri)

    def find_profile_users(self, user_name=None):
        uri = "/profile/users"
        if user_name is not None:
            uri = uri + "/" + user_name
        return self.__find(uri)

    def find_profile_groups(self, user_name, group_name=None):
        uri = "/profile/users/{}/groups".format(user_name)
        if group_name is not None:
            uri = uri + "/" + group_name
        return self.__find(uri)

    def find_profiles(self, user_name, group_name):
        if group_name is None:
            group_name = "NOGROUP"
        uri = "/profile/users/{}/groups/{}/profiles".format(user_name, group_name)
        return self.__find(uri)

    def find_linked_profiles(self, user_name, group_name, template_name):
        if group_name is None:
            group_name = "NOGROUP"
        uri = "/profile/users/{}/groups/{}/profiles/{}/configs".format(user_name, group_name, template_name)
        return self.__find(uri)

    def find_curve_groups(self, group_name=None):
        uri = "/curve/groups"
        if group_name is not None:
            uri = uri + "/" + group_name
        return self.__find(uri)

    def find_curves(self, group_name, curve_name=None):
        if group_name is None:
            group_name = 'NOGROUP'
        uri = "/curve/groups/{}/curves".format(group_name)
        if curve_name is not None:
            uri = uri + "/" + curve_name
        return self.__find(uri)

    def get_curve_validation(self, curve_group, curve_name, start_date, end_date):
        uri = "/curve/validation/groups/{}/curves/{}".format(curve_group, curve_name)
        response = self.__get_httpclient().get_resource(uri, self.__payload(start_date, end_date))
        return pd.DataFrame.from_dict(self.__data(response))

    def get_curve_group_validation(self, curve_group, start_date, end_date):
        uri = "/curve/validation/groups/{}".format(curve_group)
        response = self.__get_httpclient().get_resource(uri, self.__payload(start_date, end_date))
        return pd.DataFrame.from_dict(self.__data(response))

    def get_batch_status(self, start_date, end_date, batch_name):
        uri = "/curve/batch_status/batches/{}".format(batch_name)
        response = self.__get_httpclient().get_resource(uri, self.__payload(start_date, end_date))
        return pd.DataFrame.from_dict(self.__data(response))

    def get_batch_statuses(self, start_date, end_date, batch_type=None, batch_status=None):
        uri = "/curve/batch_status/batches"
        filters = {}
        if batch_type is not None:
            filters['type'] = batch_type
        if batch_status is not None:
            filters['status'] = batch_status

        response = self.__get_httpclient().get_resource(uri, self.__payload(start_date, end_date, None, filters))
        return pd.DataFrame.from_dict(self.__data(response))

    def get_profile(self, user_name, group_name, profile_name, config_name=None, eff_date=None, start_date=None,
                    end_date=None):
        if group_name is None:
            group_name = "NOGROUP"
        uri = "/profile/data/users/{}/groups/{}/profiles/{}".format(user_name, group_name, profile_name)
        if config_name is not None:
            uri = uri + "/configs/" + config_name

        response = self.__get_httpclient().get_resource(uri, self.__payload(start_date, end_date, None, None, eff_date))
        return pd.DataFrame.from_dict(self.__data(response))

    def get_curve(self, group_name, curve_name, start_date, end_date):
        if group_name is None:
            group_name = "NOGROUP"
        uri = "/curve/data/groups/{}/curves/{}".format(group_name, curve_name)
        response = self.__get_httpclient().get_resource(uri, self.__payload(start_date, end_date))
        return pd.DataFrame.from_dict(self.__data(response))

    def get_curves_in_group(self, group_name, start_date, end_date):
        if group_name is None:
            group_name = "NOGROUP"
        uri = "/curve/data/groups/{}".format(group_name)
        response = self.__get_httpclient().get_resource(uri, self.__payload(start_date, end_date))
        return pd.DataFrame.from_dict(self.__data(response))

    def get_timeseries_curve(self, start_date, end_date, filters=None, include_properties=None):
        return self.__curve_data("Time Series", start_date, end_date, filters, include_properties)

    def get_forward_curve(self, start_date, end_date, filters=None, include_properties=None):
        return self.__curve_data("Futures", start_date, end_date, filters, include_properties)

    def get_options_curve(self, start_date, end_date, filters=None, include_properties=None):
        return self.__curve_data("Options", start_date, end_date, filters, include_properties)

    def upload_curve_data(self, group_name, curve_name, payload, partial_update=None):
        if group_name is None:
            group_name = "NOGROUP"
        uri = "/curve/data/groups/{}/curves/{}".format(group_name, curve_name)
        if partial_update:
            response = self.__get_httpclient().patch_resource(uri, json.dumps(self.__upload_payload(payload)))
        else:
            response = self.__get_httpclient().put_resource(uri, json.dumps(self.__upload_payload(payload)))
        return self.__data(response)

    def enable_debug(self):
        if self.__httpclient is not None:
            self.__httpclient.enable_debug()

    def disable_debug(self):
        if self.__httpclient is not None:
            self.__httpclient.disable_debug()

    def close(self):
        if self.__httpclient is not None:
            self.__httpclient.get_resource('/logout')
            self.__httpclient.close()
        
    def __get_httpclient(self):
        if self.__httpclient is None:
            self.__httpclient = self.__zema_auth.authenticate()
        
        return self.__httpclient

    def __curve_data(self, curve_type, start_date, end_date, filters, include_properties):
        queries = {}
        queries['dataType'] = curve_type
        if include_properties is not None:
            if type(include_properties) == list:
                if len(include_properties) > 0:
                    queries['includeProperties'] = ",".join(include_properties)
            else:
                queries['includeProperties'] = include_properties

        if filters is not None:
            queries['filter'] = self.__curve_filter_query(filters)

        uri = "/curve/data"
        response = self.__get_httpclient().get_resource(uri, self.__payload(start_date, end_date, None, queries))
        return pd.DataFrame.from_dict(self.__data(response))

    def __payload(self, start_date=None, end_date=None, select=None, filters=None, effDate=None):
        payload = {}
        if start_date is not None:
            payload['startDate'] = str(start_date.isoformat())

        if end_date is not None:
            payload['endDate'] = str(end_date.isoformat())

        if select is not None:
            if type(select) == list:
                payload['$select'] = ",".join(select)
            else:
                payload['$select'] = select

        if filters is not None:
            for key in filters:
                t = type(filters[key])
                if t == datetime.date or t == datetime.datetime:
                    payload[key] = str(filters[key].isoformat())
                elif t == bool:
                    payload[key] = str(filters[key]).lower()
                else:
                    payload[key] = filters[key]

        if effDate is not None:
            payload['effectiveDate'] = str(effDate.isoformat())

        return payload

    def __upload_payload(self, payload):
        if payload['effectiveDate'] is not None:
            if isinstance(payload['effectiveDate'], datetime.date):
                payload['effectiveDate'] = str(payload['effectiveDate'].isoformat())

        if payload['data'] is not None:
            for data in payload['data']:
                if isinstance(data["date"], datetime.date):
                    data['date'] = str(data['date'].isoformat())
                if 'value' in data and data['value'] is not None:
                    if isinstance(data['value'], datetime.date) or isinstance(data['value'], datetime.datetime):
                        data['value'] = calendar.timegm(data['value'].timetuple()) * 1000

        return payload

    def __find(self, uri):
        response = self.__get_httpclient().get_resource(uri)
        json_object = json.loads(response.text)
        if type(json_object['data']) != list:
            return pd.DataFrame.from_dict([json_object['data']])
        else:
            return pd.DataFrame.from_dict(json_object['data'])

    def __data(self, response):
        json_object = json.loads(response.text)
        if 'data' not in json_object['data']:
            return json_object['data']
        else:
            return json_object['data']['data']

    def __curve_filter_query(self, curve_filters):
        filters = []
        for key in curve_filters:
            if key == 'properties':
                property_filters = []
                for p in curve_filters['properties']:
                    p_objects = []
                    for k in p:
                        p_objects.append(self.__single_filter(k, p[k]))
                    property_filters.append('{' + ','.join(p_objects) + '}')
                filters.append(key + ':[' + ','.join(property_filters) + ']')
            else:
                filters.append(self.__single_filter(key, curve_filters[key]))

        return '{' + ','.join(filters) + '}'

    def __single_filter(self, key, values):
        filter_clause = key + ":"

        if type(values) == list:
            quoted_values = []
            filter_clause = filter_clause + '['
            for v in values:
                quoted_values.append('"' + v + '"')
            filter_clause = filter_clause + ','.join(quoted_values) + ']'
        else:
            filter_clause = filter_clause + '"' + values + '"'

        return filter_clause



'''
ZE SOAP Web Services Proxy

Created on Jan 03, 2019

@copyright: ZE
@precondition: ZEMA Data Direct license
@version: 1.0.1
@author: ZE
'''

from lxml import etree
from zeep import Plugin
from zeep import Client

from dicttoxml import dicttoxml

class SoapProxy:
    def __init__(self,wsdl,user_name,password,client_id,proxies):
        class ExecuteReportQuery(Plugin):
            def set_xml_data(self, xml_data):
                self.xml_data = xml_data
            def egress(self, envelope, http_headers, operation, binding_options):
                r = envelope.xpath('//executeReportQueryRequest')
                if (r == None or len(r) == 0):
                    return envelope, http_headers
                p=next(r[0].iterancestors())
                p.remove(r[0])
                parser = etree.XMLParser(strip_cdata=False)
                new_request = etree.XML(self.xml_data.encode('utf-8'), parser=parser)
                p.append(new_request)
                xml_string = etree.tostring(envelope)
                parser = etree.XMLParser(strip_cdata=False)
                new_envelope = clean_request_data(etree.XML(xml_string, parser=parser))
                return new_envelope, http_headers
        
        class SendDataRequest(Plugin):
            def set_xml_data(self, xml_data):
                self.xml_data = xml_data
            def egress(self, envelope, http_headers, operation, binding_options):
                r = envelope.xpath('//request')
                if (r == None or len(r) == 0):
                    return envelope, http_headers
                p=next(r[0].iterancestors())
                p.remove(r[0])
                parser = etree.XMLParser(strip_cdata=False)
                new_request = etree.XML(self.xml_data.encode('utf-8'), parser=parser)
                p.append(new_request)
                xml_string = etree.tostring(envelope)
                parser = etree.XMLParser(strip_cdata=False)
                new_envelope = clean_request_data(etree.XML(xml_string, parser=parser))
                return new_envelope, http_headers
            
        self.wsdl=wsdl
        self.plugin = ExecuteReportQuery()
        self.plugin2 = SendDataRequest()
        self.client = Client(wsdl=wsdl, plugins=[self.plugin, self.plugin2])
                 
        self.token=self.client.service.getAccessTokenForAccount(user_name,password,client_id)
        if proxies is not None:
            self.client.transport.session.proxies = proxies

    def get_token(self):
        return self.token
    def execute_profile(self, data):
        pass
    def execute_report_query(self,data_type,data):
        pass
    def find_curves(self,data):
        pass
    def retrieve_curve_data_beans(self,data):
        pass
    def get_forward_curve_data(self,data):
        pass
    def get_time_series_curve_data(self,data):
        pass
    def get_options_curve_data(self,data):
        pass
    def retrieve_observations(self,data):
        pass
    def send_data_request(self,data_type,data,fmt):
        pass
    def insert_update_curve_data(self,data):
        pass
    def close(self):
        pass

def check_token(token):
    assert(token != None), "The proxy object has been detroyed"

def execute_profile(self, data):
    check_token(self.token)
    return self.client.service.executeProfile(self.token,data)

def retrieve_curve_data_beans(self, data):
    check_token(self.token)
    return self.client.service.retrieveCurveDataBeans(self.token,data)

def find_curves(self, data):
    check_token(self.token)
    return self.client.service.findCurves(self.token,data)

def get_forward_curve_data(self, data):
    check_token(self.token)
    return self.client.service.getForwardCurveData(self.token,data)

def get_time_series_curve_data(self, data):
    check_token(self.token)
    return self.client.service.getTimeSeriesCurveData(self.token,data)

def get_options_curve_data(self, data):
    check_token(self.token)
    return self.client.service.getOptionsCurveData(self.token,data)

def retrieve_observations(self, datasource, report):
    check_token(self.token)
    return self.client.service.retrieveObservations(self.token,datasource,report)

def insert_update_curve_data(self, data):
    check_token(self.token)
    return self.client.service.insertUpdateCurveData(self.token,data)

def close(self):
    self.client.service.logout(self.token)
    self.token = None

def execute_report_query(self, data_type, data):
    TIME_SERIES_TYPE="timeSeriesRequest"
    ADVANCED_TYPE="advancedQueryRequest"
    if data_type=='Time Series':
        request_type=TIME_SERIES_TYPE
    elif data_type=='Data Sheet':
        request_type=ADVANCED_TYPE
    else:
        print("Usage: execute_report_query('Time Series' or 'Data Sheet', data)")
        return

    xml_data = dicttoxml(data,attr_type=False,custom_root='executeReportQueryRequest').decode()
    xml_data = xml_data.replace("<dataRequest>",'<dataRequest xmlns:ns6="datarequest.datadirect.ze.com" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xsi:type="ns6:'+request_type+'">') 
    self.plugin.set_xml_data(xml_data)

    result = self.client.service.executeReportQuery(self.token, {'useDisplayName': '',
                                                   'dataRequest':{}
                                                   })
    return result

def send_data_request(self, data_type, data, fmt):
    TIME_SERIES_TYPE="timeSeriesRequest"
    ADVANCED_TYPE="advancedQueryRequest"
    if data_type=='Time Series':
        request_type=TIME_SERIES_TYPE
    elif data_type=='Data Sheet':
        request_type=ADVANCED_TYPE
    else:
        print("Usage: sendDataRequest('Time Series' or 'Data Sheet', data)")
        return

    xml_data = dicttoxml(data,attr_type=False,custom_root='request').decode()
    xml_data = xml_data.replace("<request>",'<request xmlns:ns6="datarequest.datadirect.ze.com" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xsi:type="ns6:'+request_type+'">') 
    self.plugin2.set_xml_data(xml_data)

    result = self.client.service.sendDataRequest(self.token, 
                                                 {},
                                                 fmt
                                                )
    return result

def clean_request_data(new_envelope):
    for elem in new_envelope.findall(".//item"):
        parent = elem.find('..')
        if parent is not None:
            elem.tag = parent.tag
            grand = parent.find('..')
            grand.append(elem)
            
    for elem in new_envelope.xpath(".//*[not(node())]"):
        elem.getparent().remove(elem)
        
    return new_envelope

SoapProxy.execute_profile=execute_profile
SoapProxy.retrieve_curve_data_beans=retrieve_curve_data_beans
SoapProxy.find_curves=find_curves
SoapProxy.execute_report_query=execute_report_query
SoapProxy.get_forward_curve_data=get_forward_curve_data
SoapProxy.get_time_series_curve_data=get_time_series_curve_data
SoapProxy.get_options_curve_data=get_options_curve_data
SoapProxy.retrieve_observations=retrieve_observations
SoapProxy.send_data_request = send_data_request
SoapProxy.insert_update_curve_data=insert_update_curve_data
SoapProxy.close=close




'''
Created on May 8, 2019
Updated on April 12, 2022

@author: ZE
'''

from distutils.core import setup

test_dependencies = [
        'pytest-cov',
        'pytest',
        'configparser',
        'numpy',
        'tabulate',
        'validate_email'
]

setup(
    name='zeclient',
    version='1.1.3',
    author='ZE',
    author_email='support@ze.com',
    packages=['zeclient', 'zeclient.oauth2'],
    url='http://www.ze.com',
    license='LICENSE.txt',
    description='Python APIs for ZE Web Services.',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    python_requires='>=3.5',
    install_requires=[
        'requests',
        'lxml',
        'dicttoxml',
        'pandas',
        'zeep',
        'beautifulsoup4'
    ],
    tests_require=test_dependencies,
    extras_require={'test': test_dependencies}
)



class AdfsAuthorizationCode:
    def __init__(self, ze_http_client, dd_url, username, password, verify_ssl=True) -> None:
        self.dd_authorize_url = dd_url + "/oauth2/authorize"
        self.ze_http_client = ze_http_client
        self.verify_ssl = verify_ssl
        self.username = username
        self.password = password
        pass

    def authenticate(self):
        session = self.ze_http_client.session
        response = self.ze_http_client.get(self.dd_authorize_url)

        adfs_authorize_url = response.url

        payload = {'UserName': self.username,
                   'Password': self.password,
                   'AuthMethod': 'FormsAuthentication'}

        headers = {
            'Content-Type': 'application/x-www-form-urlencoded'
        }
        response = self.ze_http_client.post(adfs_authorize_url, data=payload, proxies=self.ze_http_client.proxies,
                                            headers=headers)



import base64


class OktaResourceOwnerPassword:
    def __init__(self, ze_http_client, okta_issuer_url, username, password, idp_client_id, idp_client_secret,
                 verify_ssl=True) -> None:
        self.okta_token_url = okta_issuer_url + "/v1/token"
        self.ze_http_client = ze_http_client
        self.verify_ssl = verify_ssl
        self.username = username
        self.password = password
        self.idp_client_id = idp_client_id
        self.idp_client_secret = idp_client_secret
        pass

    def get_access_token(self):
        session = self.ze_http_client.session

        payload = {'username': self.username,
                   'password': self.password,
                   'grant_type': 'password',
                   'scope': 'openid'}

        code = f'{self.idp_client_id}:{self.idp_client_secret}'
        encoded = base64.b64encode(code.encode('ascii')).decode('ascii')

        headers = {
            'authorization': f'Basic {encoded}',
            'content-type': 'application/x-www-form-urlencoded'
        }
        response = self.ze_http_client.post(self.okta_token_url, data=payload, proxies=self.ze_http_client.proxies,
                                            headers=headers)

        tokens = response.json()
        return tokens["access_token"]
